@@ -8,6 +8,7 @@ Example usage:
 ...prints list of available time series
 """
 
+from archimedes.auth import archimedes_auth
 from archimedes.configuration import constants  # config
 from archimedes.data.api import (
     get,
@@ -37,6 +38,7 @@ from archimedes.utils import compact_print, full_print, environ
 
 
 __all__ = [
+    "archimedes_auth",
     "compact_print",
     "full_print",
     "environ",
