@@ -9,14 +9,10 @@ from typing import Dict, List
 import numpy as np
 import pandas as pd
 import requests
-from requests.exceptions import HTTPError
-
-# TODO when requests v2.27.0+ is released, we can use JSONDecodeError directly from requests.exceptions
-# TODO See https://github.com/psf/requests/pull/5856 for more details
-from simplejson import JSONDecodeError
+from requests.exceptions import HTTPError, JSONDecodeError
 
 import archimedes
-from archimedes.auth import archimedes_auth
+from archimedes.auth import archimedes_auth, NoneAuth
 from archimedes.configuration import api_config
 
 API_URL = api_config.url
@@ -28,6 +24,8 @@ def get(
     price_areas: List[str] = None,
     start: str = None,
     end: str = None,
+    *,
+    access_token: str = None,
 ):
     """Get any number of time series.
 
@@ -55,6 +53,7 @@ def get(
         price_areas (List[str], optional): The price areas to pick, all price areas if None. Defaults to None.
         start (str, optional): The first datetime to fetch (inclusive). Returns all if None. Defaults to None.
         end (str, optional): The last datetime to fetch (exclusive). Returns all if None. Defaults to None.
+        access_token (str, optional): None - access token for the API
 
     Returns:
         DataFrame with all the time series data
@@ -90,6 +89,7 @@ def get(
 
     data = _make_api_request(
         f"{API_URL}/v{API_VERSION}/data/get",
+        access_token=access_token,
         params=query,
     )
 
@@ -116,6 +116,8 @@ def get(
 def get_latest(
     series_ids: List[str],
     price_areas: List[str] = None,
+    *,
+    access_token: str = None,
 ):
     """Get the most recent data for any number of time series.
 
@@ -143,6 +145,7 @@ def get_latest(
     Args:
         series_ids (List[str]): The series ids to get.
         price_areas (List[str], optional): The price areas to pick, all price areas if None. Defaults to None.
+        access_token (str, optional): None - access token for the API
 
     Returns:
         DataFrame with the latest time series data
@@ -161,6 +164,7 @@ def get_latest(
         price_areas=price_areas,
         start=start_dt.isoformat(),
         end=end_dt.isoformat(),
+        access_token=access_token,
     )
 
     return df
@@ -171,6 +175,8 @@ def get_predictions(
     price_areas: List[str] = None,
     start: str = None,
     end: str = None,
+    *,
+    access_token: str = None,
 ) -> List:
     """Get any number of predictions
 
@@ -193,6 +199,7 @@ def get_predictions(
             The first datetime to fetch (inclusive). Returns all if None. Defaults to None.
         end (str, optional):
             The last datetime to fetch (exclusive). Returns all if None. Defaults to None.
+        access_token (str, optional): None - access token for the API
 
     Returns:
         DataFrame with all the prediction data
@@ -223,6 +230,7 @@ def get_predictions(
 
     data = _make_api_request(
         f"{API_URL}/v{API_VERSION}/data/get_predictions",
+        access_token=access_token,
         params=query,
     )
 
@@ -234,7 +242,7 @@ def get_predictions(
     return pd.DataFrame.from_dict(data)
 
 
-def list_series_price_areas(series_id: str):
+def list_series_price_areas(series_id: str, *, access_token: str = None):
     """
     Retrieve all of the price_areas which are available for the specified data series
 
@@ -254,14 +262,14 @@ def list_series_price_areas(series_id: str):
         "series_id": series_id,
     }
     data = _make_api_request(
-        f"{API_URL}/v{API_VERSION}/data/list_series_price_areas", params=query
+        f"{API_URL}/v{API_VERSION}/data/list_series_price_areas", access_token=access_token, params=query
     )
 
     price_area_df = pd.DataFrame.from_dict(data)
     return price_area_df
 
 
-def list_ids(sort: bool = False):
+def list_ids(sort: bool = False, *, access_token: str = None):
     """List all the series ids available.
 
     Example:
@@ -275,12 +283,14 @@ def list_ids(sort: bool = False):
 
     Args:
         sort (bool): False - return all series in one dataframe column, True - order dataframe by data-origin
+        access_token (str, optional): None - access token for the API
 
     Returns:
         DataFrame with all available list_ids
     """
     data = _make_api_request(
         f"{API_URL}/v{API_VERSION}/data/list_ids",
+        access_token=access_token
     )
 
     series_df = pd.DataFrame.from_dict(data)
@@ -309,7 +319,7 @@ def list_ids(sort: bool = False):
     return sorted_df.copy()
 
 
-def list_prediction_ids():
+def list_prediction_ids(*, access_token: str = None):
     """List all the prediction series ids available.
 
     Example:
@@ -322,12 +332,12 @@ def list_prediction_ids():
         23                         PX/rk-naive
     """
 
-    data = _make_api_request(f"{API_URL}/v{API_VERSION}/data/list_prediction_ids")
+    data = _make_api_request(f"{API_URL}/v{API_VERSION}/data/list_prediction_ids", access_token=access_token)
 
     return pd.DataFrame.from_dict(data)
 
 
-def get_predictions_ref_dts(prediction_id: str = None):
+def get_predictions_ref_dts(prediction_id: str = None, *, access_token: str = None):
     """Get which ref_dts are available.
 
     ref_dt == prediction_build_dt
@@ -335,6 +345,7 @@ def get_predictions_ref_dts(prediction_id: str = None):
 
     Args:
         prediction_id (str): The series id to get the reference dts for. If None, get ref_dts for all prediction_ids.
+        access_token (str, optional): Access token for the API
 
     Returns:
         DataFrame with all ref_dts
@@ -346,14 +357,18 @@ def get_predictions_ref_dts(prediction_id: str = None):
 
     data = _make_api_request(
         f"{API_URL}/v{API_VERSION}/data/get_predictions_ref_dts",
+        access_token=access_token,
         params=query,
     )
 
     return pd.DataFrame.from_dict(data)
 
 
-def _make_api_request(url, method="GET", *args, **kwargs):
-    access_token = archimedes_auth.get_access_token_silent()
+def _make_api_request(url, method="GET", access_token=None, *args, **kwargs):
+    if access_token is None:
+        if archimedes_auth is None:
+            raise NoneAuth("access_token parameter must be passed when using USE_WEB_AUTHENTICATION")
+        access_token = archimedes_auth.get_access_token_silent()
 
     headers = kwargs.pop("headers", None)
     if headers is None:
@@ -365,7 +380,13 @@ def _make_api_request(url, method="GET", *args, **kwargs):
 
     if r.status_code not in [HTTPStatus.OK, HTTPStatus.CREATED]:
         try:
-            error_message = r.json().get("detail")
+            response_json = r.json()
+            if "message" in response_json:
+                error_message = response_json.get("message")
+            elif "detail" in response_json:
+                error_message = response_json.get("detail")
+            else:
+                error_message = json.dumps(response_json)
         except JSONDecodeError:
             error_message = r.content
         raise HTTPError(f"API Error: {error_message}")
@@ -379,6 +400,8 @@ def store_prediction(
     ref_dt: pd.Timestamp,
     run_dt: pd.Timestamp,
     data: Dict,
+    *,
+    access_token: str = None,
 ):
     """Store a prediction
 
@@ -414,6 +437,7 @@ def store_prediction(
         method="POST",
         data=json.dumps(payload),
         headers={"Content-Type": "application/json"},
+        access_token=access_token,
     )
 
     return ret.get("success", False)
@@ -424,19 +448,24 @@ def forecast_list_ref_times(
     start: pd.Timestamp = None,
     end: pd.Timestamp = None,
     limit: int = None,
+    *,
+    access_token: str = None,
 ):
     """
     List all forecast reference times (the times that the forecast was generated)
 
     Args:
-        series_id: The ID of the data series to find all the forecast reference times (the time the forecast was
+        series_id (str): The ID of the data series to find all the forecast reference times (the time the forecast was
             generated). Retrieve the complete list of series (both forecasts and observations) using the list_ids
             resource.
-        start: The first datetime to fetch (inclusive). Returns all if not set. Should be specified in ISO 8601 format
+        start (pd.Timestamp, optional): The first datetime to fetch (inclusive). Returns all if not set. Should be
+            specified in ISO 8601 format.
             (eg - '2021-11-29T06:00:00+00:00')
-        end: The last datetime to fetch (exclusive). Returns all if not set. Should be specified in ISO 8601 format
+        end (pd.Timestamp, optional): The last datetime to fetch (exclusive). Returns all if not set. Should be
+            specified in ISO 8601 format.
             (eg - '2021-11-30T06:00:00+00:00')
-        limit: Limit the output to a specific number of entries. No limit if not specified.
+        limit (int, optional): Limit the output to a specific number of entries. No limit if not specified.
+        access_token (str, optional): Access token for the API
 
     Example:
         >>> start = pd.Timestamp("2022-01-09T06:00:00+00:00")
@@ -460,7 +489,7 @@ def forecast_list_ref_times(
     }
 
     data = _make_api_request(
-        f"{API_URL}/v{API_VERSION}/forecast/list_ref_times", params=query
+        f"{API_URL}/v{API_VERSION}/forecast/list_ref_times", access_token=access_token, params=query
     )
 
     return pd.DataFrame.from_dict(data)
@@ -472,6 +501,8 @@ def forecast_diff(
     price_areas: List[str] = None,
     ref_time1: pd.Timestamp = None,
     ref_time2: pd.Timestamp = None,
+    *,
+    access_token: str = None,
 ):
     """
     Get the difference between two different forecasts.
@@ -495,6 +526,7 @@ def forecast_diff(
             format (eg - '2021-11-29T06:00:00+00:00').
         ref_time2: Specify another of the two timestamps for when a forecast was created. Should be specified in ISO
             8601 format (eg - '2021-11-29T06:00:00+00:00').
+        access_token (str, optional): Access token for the API
 
     Returns:
         Dataframe with the diff of the two forecasts
@@ -511,7 +543,7 @@ def forecast_diff(
         "ref_time2": ref_time2,
     }
 
-    data = _make_api_request(f"{API_URL}/v{API_VERSION}/forecast/diff", params=query)
+    data = _make_api_request(f"{API_URL}/v{API_VERSION}/forecast/diff", access_token=access_token, params=query)
 
     return pd.DataFrame.from_dict(data)
 
@@ -523,6 +555,8 @@ def forecast_get_by_ref_time_interval(
     end: pd.Timestamp = None,
     forecast_interval: int = 24,
     day_ahead_hour: int = None,
+    *,
+    access_token: str = None,
 ):
     """
     Get a single forecast value for every hour. The value should be from the forecast that was generated at least
@@ -545,6 +579,7 @@ def forecast_get_by_ref_time_interval(
             ('forecast_interval' will be set to 24). For example, if set to '12' (noon CET), the values shown for every
             hour of a specific day will be fetched from the most recent forecast generated before noon (most likely
             11am) on the previous day.
+        access_token (str, optional): Access token for the API
 
     Returns:
         Dataframe with the forecasted values
@@ -559,7 +594,7 @@ def forecast_get_by_ref_time_interval(
     }
 
     data = _make_api_request(
-        f"{API_URL}/v{API_VERSION}/forecast/get_by_ref_time_interval", params=query
+        f"{API_URL}/v{API_VERSION}/forecast/get_by_ref_time_interval", access_token=access_token, params=query
     )
 
     return pd.DataFrame.from_dict(data)
